/* 0.02 - Copyright (c) 2015, Martin Albert <ma@debian.org>
 *
 * Permission is hereby granted, free of charge, to any person 
 * obtaining a copy of this software and associated documentation 
 * files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, 
 * merge, publish, distribute, sublicense, and/or sell copies of 
 * the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be 
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE 
 * OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <stdint.h>
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "screendef.h"
#include "vernr.h"
#include "rusage.h"
#include "thrdef.h"
#include "hwaredef.h"

#include "sdrdef.h"
#include "conf.h"
#include "options.h"

/* only ONE HackRF supported and only HackRF ONE */
#define MAX_HACKRF_DEVICES 1
#define ADJ_HACKRF_XTAL 1

typedef enum hackrf_error {
	HACKRF_SUCCESS = 0,
	HACKRF_TRUE = 1,
	HACKRF_ERROR_INVALID_PARAM = -2,
	HACKRF_ERROR_NOT_FOUND = -5,
	HACKRF_ERROR_BUSY = -6,
	HACKRF_ERROR_NO_MEM = -11,
	HACKRF_ERROR_LIBUSB = -1000,
	HACKRF_ERROR_THREAD = -1001,
	HACKRF_ERROR_STREAMING_THREAD_ERR = -1002,
	HACKRF_ERROR_STREAMING_STOPPED = -1003,
	HACKRF_ERROR_STREAMING_EXIT_CALLED = -1004,
	HACKRF_ERROR_OTHER = -9999,
} hackrf_error;

typedef struct hackrf_device hackrf_device;

typedef struct {
	hackrf_device* device;
	uint8_t* buffer;
	int buffer_length;
	int valid_length;
	void* rx_ctx;
	void* tx_ctx;
} hackrf_transfer_t, hackrf_transfer;

typedef struct {
	uint32_t part_id[2];
	uint32_t serial_no[4];
} hackrf_read_partid_serialno_t;

typedef int (*hackrf_sample_block_cb_fn)(hackrf_transfer* transfer);

struct hackrf_device {
	void* usb_device;
	struct libusb_transfer** transfers;
	hackrf_sample_block_cb_fn callback;
	volatile int transfer_thread_started; /* volatile shared between threads (read only) */
	pthread_t transfer_thread;
	uint32_t transfer_count;
	uint32_t buffer_size;
	volatile int streaming; /* volatile shared between threads (read only) */
	void* rx_ctx;
	void* tx_ctx;
};


typedef int (*p_hackrf_init)(void);
typedef int (*p_hackrf_exit)(void);
typedef int (*p_hackrf_open)(struct hackrf_device** device);
typedef char* (*p_hackrf_error_name)(int code);
typedef int (*p_hackrf_board_partid_serialno_read)(struct hackrf_device* 
                              device, hackrf_read_partid_serialno_t* id);
typedef int (*p_hackrf_version_string_read)(struct hackrf_device* device,
                                           char* version, uint8_t length);
typedef int (*p_hackrf_board_id_read)(struct hackrf_device* device, uint8_t* value);
typedef const char* (*p_hackrf_board_id_name)(uint8_t board_id);
typedef int (*p_hackrf_close)(struct hackrf_device* device);
typedef int (*p_hackrf_set_baseband_filter_bandwidth)(struct hackrf_device* device, 
                                                          uint32_t bandwidth_hz);
typedef int (*p_hackrf_set_samplerate)(struct hackrf_device* device,
                                                        double rate);
typedef int (*p_hackrf_start_rx)(struct hackrf_device* device, 
                          hackrf_sample_block_cb_fn callback, void* rx_ctx);
typedef int (*p_hackrf_set_freq)(struct hackrf_device* device, 
                                                          uint64_t freq_hz);
typedef int (*p_hackrf_set_lna_gain)(struct hackrf_device* device, 
                                                          uint32_t value);
typedef int (*p_hackrf_set_vga_gain)(struct hackrf_device* device, 
                                                          uint32_t value);
typedef int (*p_hackrf_set_amp_enable)(struct hackrf_device* device,
                                                          uint8_t value);
typedef int (*p_hackrf_set_antenna_enable)(struct hackrf_device* dev,
/* Antenna port power */                                   uint8_t x);

p_hackrf_init hackrf_init;
p_hackrf_exit hackrf_exit;
p_hackrf_open hackrf_open;
p_hackrf_error_name hackrf_error_name;
p_hackrf_board_partid_serialno_read hackrf_board_partid_serialno_read;
p_hackrf_version_string_read hackrf_version_string_read;
p_hackrf_board_id_read hackrf_board_id_read;
p_hackrf_board_id_name hackrf_board_id_name;
p_hackrf_close hackrf_close;
p_hackrf_set_baseband_filter_bandwidth hackrf_set_baseband_filter_bandwidth;
p_hackrf_set_samplerate hackrf_set_samplerate;
p_hackrf_start_rx hackrf_start_rx;
p_hackrf_set_freq hackrf_set_freq;
p_hackrf_set_lna_gain hackrf_set_lna_gain;
p_hackrf_set_vga_gain hackrf_set_vga_gain;
p_hackrf_set_amp_enable hackrf_set_amp_enable;
p_hackrf_set_antenna_enable hackrf_set_antenna_enable;


// Structure for HackRf parameters
typedef struct {
unsigned int ratenum;
unsigned int sernum1;
unsigned int sernum2;
int freq_adjust;
unsigned int gain_mode;
unsigned int bandwidth_no;
int bias_t;
int check;
}P_HACKRF;
#define MAX_HACKRF_PARM 8

/*#define MAX_HACKRF_SR 5*/
#define MAX_HACKRF_SR 8
unsigned int hackrf_sr[MAX_HACKRF_SR]=
  {20000000,16000000,12500000,10000000,8000000,5000000,3500000,2000000};

#define MAX_HACKRF_BW 16
float hackrf_bw[MAX_HACKRF_BW]={28.0F,24.0F,20.0F,15.0F,14.0F,12.0F,10.0F,9.0F,
                                 8.0F,7.0F,6.0F,5.5F,5.0F,3.5F,2.5F,1.75F};
#define TOMD(a) ((a)/1000000.)

P_HACKRF hackrf;
char *hackrf_parfil_name="par_hackrf";
struct hackrf_device *dev_hackrf;
int hackrf_library_flag;
int cancel_flag;

char *hackrf_parm_text[MAX_HACKRF_PARM]={"Sampling speed",
                                          "Serno1",
                                          "Serno2",
                                          "Freq adjust",
                                          "Gain Mode",
                                          "Bandwidth No",
                                          "Bias T",
                                          "Check"};

#if(OSNUM == OSNUM_WINDOWS)
#include <windows.h>
HANDLE hackrf_libhandle;

void load_hackrf_library(void)
{
int info;
if(hackrf_library_flag)return;
info=0;
hackrf_libhandle=LoadLibrary("hackrf.dll");
if(!hackrf_libhandle)goto hackrf_load_error;
info=1;
hackrf_init=(p_hackrf_init)GetProcAddress(hackrf_libhandle, "hackrf_init");
if(!hackrf_init)goto hackrf_sym_error;
hackrf_exit=(p_hackrf_exit)GetProcAddress(hackrf_libhandle, "hackrf_exit");
if(!hackrf_exit)goto hackrf_sym_error;
hackrf_open=(p_hackrf_open)GetProcAddress(hackrf_libhandle, "hackrf_open");
if(!hackrf_open)goto hackrf_sym_error;
hackrf_error_name=(p_hackrf_error_name)GetProcAddress(
                                 hackrf_libhandle, "hackrf_error_name");
if(!hackrf_error_name)goto hackrf_sym_error;
hackrf_board_partid_serialno_read=(p_hackrf_board_partid_serialno_read)
                                   GetProcAddress(hackrf_libhandle,
                                   "hackrf_board_partid_serialno_read");
if(!hackrf_board_partid_serialno_read)goto hackrf_sym_error;
hackrf_version_string_read=(p_hackrf_version_string_read)GetProcAddress(
                        hackrf_libhandle, "hackrf_version_string_read");
if(!hackrf_version_string_read)goto hackrf_sym_error;
hackrf_board_id_read=(p_hackrf_board_id_read)GetProcAddress(
                              hackrf_libhandle, "hackrf_board_id_read");
if(!hackrf_board_id_read)goto hackrf_sym_error;
hackrf_board_id_name=(p_hackrf_board_id_name)GetProcAddress(
                              hackrf_libhandle, "hackrf_board_id_name");
if(!hackrf_board_id_name)goto hackrf_sym_error;
hackrf_close=(p_hackrf_close)GetProcAddress(hackrf_libhandle, "hackrf_close");
if(!hackrf_close)goto hackrf_sym_error;
hackrf_set_baseband_filter_bandwidth=(p_hackrf_set_baseband_filter_bandwidth)           
                                             GetProcAddress(
              hackrf_libhandle, "hackrf_set_baseband_filter_bandwidth");
if(!hackrf_set_baseband_filter_bandwidth)goto hackrf_sym_error;
hackrf_set_samplerate=(p_hackrf_set_samplerate)GetProcAddress(hackrf_libhandle, 
                                                   "hackrf_set_sample_rate");
if(!hackrf_set_samplerate)goto hackrf_sym_error;
hackrf_set_antenna_enable=(p_hackrf_set_antenna_enable)GetProcAddress(hackrf_libhandle, 
                                                     "hackrf_set_antenna_enable");
if(!hackrf_set_antenna_enable)goto hackrf_sym_error;*/
hackrf_start_rx=(p_hackrf_start_rx)GetProcAddress(hackrf_libhandle, 
                                                       "hackrf_start_rx");
if(!hackrf_start_rx)goto hackrf_sym_error;
hackrf_set_freq=(p_hackrf_set_freq)GetProcAddress(hackrf_libhandle, 
                                                       "hackrf_set_freq");
if(!hackrf_set_freq)goto hackrf_sym_error;
hackrf_set_lna_gain=(p_hackrf_set_lna_gain)GetProcAddress(hackrf_libhandle, 
                                                   "hackrf_set_lna_gain");
if(!hackrf_set_lna_gain)goto hackrf_sym_error;
hackrf_set_amp_enable=(p_hackrf_set_amp_enable)GetProcAddress(hackrf_libhandle, 
                                                 "hackrf_set_amp_enable");
if(!hackrf_set_amp_enable)goto hackrf_sym_error;
hackrf_set_vga_gain=(p_hackrf_set_vga_gain)GetProcAddress(hackrf_libhandle, 
                                                  "hackrf_set_vga_gain");
if(!hackrf_set_vga_gain)goto hackrf_sym_error;
hackrf_library_flag=TRUE;
return;
hackrf_sym_error:;
FreeLibrary(hackrf_libhandle);
hackrf_load_error:;
library_error_screen("hackrf.dll",info);
}

void unload_hackrf_library(void)
{
if(!hackrf_library_flag)return;
FreeLibrary(hackrf_libhandle);
hackrf_library_flag=FALSE;
}
#endif

#if(OSNUM == OSNUM_LINUX)
#include <dlfcn.h>
static void *hackrf_libhandle=NULL;

void load_hackrf_library(void)
{ 
int info;
info=0;
if(hackrf_library_flag)return;
hackrf_libhandle=dlopen(HACKRF_LIBNAME, RTLD_LAZY);
if(!hackrf_libhandle)goto hackrf_load_error;
info=1;
hackrf_init=(p_hackrf_init)dlsym(hackrf_libhandle, "hackrf_init");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_exit=(p_hackrf_exit)dlsym(hackrf_libhandle, "hackrf_exit");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_open=(p_hackrf_open)dlsym(hackrf_libhandle, "hackrf_open");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_error_name=(p_hackrf_error_name)dlsym(
                                   hackrf_libhandle, "hackrf_error_name");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_board_partid_serialno_read=(p_hackrf_board_partid_serialno_read) 
               dlsym(hackrf_libhandle, "hackrf_board_partid_serialno_read");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_version_string_read=(p_hackrf_version_string_read)
                      dlsym(hackrf_libhandle, "hackrf_version_string_read");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_board_id_read=(p_hackrf_board_id_read)
                            dlsym(hackrf_libhandle, "hackrf_board_id_read");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_board_id_name=(p_hackrf_board_id_name)
                            dlsym(hackrf_libhandle, "hackrf_board_id_name");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_close=(p_hackrf_close)dlsym(hackrf_libhandle, "hackrf_close");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_set_baseband_filter_bandwidth=(p_hackrf_set_baseband_filter_bandwidth)dlsym(hackrf_libhandle, 
                                                   "hackrf_set_baseband_filter_bandwidth");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_set_samplerate=(p_hackrf_set_samplerate)dlsym(hackrf_libhandle, 
                                                   "hackrf_set_sample_rate");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_set_antenna_enable=(p_hackrf_set_antenna_enable)dlsym(hackrf_libhandle, 
                                                     "hackrf_set_antenna_enable");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_start_rx=(p_hackrf_start_rx)dlsym(hackrf_libhandle, 
                                                       "hackrf_start_rx");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_set_freq=(p_hackrf_set_freq)dlsym(hackrf_libhandle, 
                                                       "hackrf_set_freq");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_set_lna_gain=(p_hackrf_set_lna_gain)dlsym(hackrf_libhandle, 
                                                   "hackrf_set_lna_gain");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_set_amp_enable=(p_hackrf_set_amp_enable)dlsym(hackrf_libhandle, 
                                                 "hackrf_set_amp_enable");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_set_vga_gain=(p_hackrf_set_vga_gain)dlsym(hackrf_libhandle, 
                                                  "hackrf_set_vga_gain");
if(dlerror() != 0)goto hackrf_sym_error;
hackrf_library_flag=TRUE;
return;
hackrf_sym_error:;
dlclose(hackrf_libhandle);
hackrf_library_flag=FALSE;
hackrf_load_error:;
library_error_screen(HACKRF_LIBNAME,info);
}

void unload_hackrf_library(void)
{
if(!hackrf_library_flag)return;
dlclose(hackrf_libhandle);
hackrf_library_flag=FALSE;
}
#endif


static int hackrf_callback(hackrf_transfer_t* transfer)
{
unsigned int i, samples;
short int *iz;
uint8_t *iq;
if(cancel_flag)return -1;
iz=(short int*)&timf1_char[timf1p_sdr];
iq=transfer->buffer;
samples=transfer->valid_length;
for(i=0; i<samples; i+=2)
  {
/*iz[i]=((short int)(iq[i]))<<8;
  iz[i+1]=((short int)(iq[i+1]))<<8;*/
  *iz++ = *iq++ << 8;
  *iz++ = *iq++ << 8;
  }
timf1p_sdr=(timf1p_sdr+2*samples)&timf1_bytemask;
if( ((timf1p_sdr-timf1p_pa+timf1_bytes)&timf1_bytemask) >= 
                                                     snd[RXAD].block_bytes)
  {
  lir_set_event(EVENT_HWARE1_RXREADY);
  }
return 0;
}

void hackrf_starter(void)
{
hackrf_error res;
cancel_flag=FALSE;
while(thread_status_flag[THREAD_HACKRF_INPUT]!=THRFLAG_ACTIVE)
  {
  lir_sleep(10000);
  if(kill_all_flag)return;
  }
thread_status_flag[THREAD_HACKRF_STARTER]=THRFLAG_ACTIVE;
lir_sched_yield();
res = hackrf_start_rx(dev_hackrf, hackrf_callback, NULL);
if(res != HACKRF_SUCCESS )lirerr(564989);
}


uint8_t hackrf_vga_gain0[]={62,58,56,52,50,46,44,40,38,34,32,
                            28,26,22,20,16,14,10, 8, 4, 2, 0};
uint8_t hackrf_lna_gain0[]={40,40,40,32,32,32,32,24,24,24,24,
                            24,16,16,16,16, 8, 8, 8, 8, 0, 0};
uint8_t hackrf_vga_gain1[]={62,60,58,56,54,50,48,46,44,42,38,
                            36,34,30,26,22,20,16,12, 8, 4, 0};
uint8_t hackrf_lna_gain1[]={40,40,40,32,32,32,32,24,24,24,24,
                            24,16,16,16,16, 8, 8, 8, 8, 0, 0};

void set_hackrf_att(void)
{
unsigned int gain_no=-fg.gain/4;
uint8_t vga_gain;
uint8_t lna_gain;
uint8_t amp;
if(hackrf.gain_mode == 0)
  {
  vga_gain=hackrf_vga_gain0[gain_no];
  lna_gain=hackrf_lna_gain0[gain_no];
  amp=0;
  }
else
  {
  vga_gain=hackrf_vga_gain1[gain_no];
  lna_gain=hackrf_lna_gain1[gain_no];
  amp=1;
  }

{
char s[80];
if (amp) sprintf(s,"VGA=%.2d LNA=%.2d",vga_gain,lna_gain);
else     sprintf(s,"vga=%.2d lna=%.2d",vga_gain,lna_gain);
lir_text(0,11,s);
}

hackrf_set_vga_gain(dev_hackrf, vga_gain);
hackrf_set_lna_gain(dev_hackrf, lna_gain);
hackrf_set_amp_enable(dev_hackrf, amp);
}

void set_hackrf_frequency(void)
{
unsigned int freq_hz;
freq_hz=(int)(1000000.F*fg.passband_center);
hackrf_set_freq(dev_hackrf, freq_hz);
}


void hackrf_input(void)
{
struct hackrf_device* devices[MAX_HACKRF_DEVICES];
hackrf_read_partid_serialno_t id;
hackrf_error res;
float t1;
int errcod;
#if RUSAGE_OLD == TRUE
int local_workload_counter;
#endif
int i, no_of_hackrf;
int rxin_local_workload_reset;
double dt1, read_start_time, total_reads;
int timing_loop_counter,timing_loop_counter_max,initial_skip_flag;
int local_att_counter;
int local_nco_counter;
char s[80];
lir_init_event(EVENT_HWARE1_RXREADY);
#if OSNUM == OSNUM_LINUX
clear_thread_times(THREAD_HACKRF_INPUT);
#endif
#if RUSAGE_OLD == TRUE
local_workload_counter=workload_counter;
#endif
local_att_counter=sdr_att_counter;
local_nco_counter=sdr_nco_counter;
dt1=current_time();
i=read_sdrpar(hackrf_parfil_name, MAX_HACKRF_PARM, 
                                  hackrf_parm_text, (int*)((void*)&hackrf));
errcod=1415;
if(i != 0 || 
   hackrf.check != HACKRF_PAR_VERNR ||
   hackrf.bandwidth_no >= MAX_HACKRF_BW)goto hackrf_init_error_exit;
t1=hackrf.ratenum;
t1=((rint)((t1*(100000000L+(double)hackrf.freq_adjust))/100000000L));
if(ui.rx_ad_speed != (int)t1)
  {
  errcod=1415;
  goto hackrf_init_error_exit;
  }
timf1p_sdr=timf1p_pa;

load_usb1_library(FALSE);
errcod=1402;
if(libusb1_library_flag == FALSE)goto hackrf_init_error_exit;
load_hackrf_library();
errcod=1414;
if(!hackrf_library_flag)goto hackrf_init_error_exit1;
errcod=23723;
no_of_hackrf=0;
hackrf_init();/*!!!*/
while(!kill_all_flag)
  {
  res=HACKRF_SUCCESS;
  while( res == HACKRF_SUCCESS && no_of_hackrf < MAX_HACKRF_DEVICES)
    {
    devices[no_of_hackrf]=NULL;
    res=hackrf_open(&devices[no_of_hackrf]);
    if(res == HACKRF_SUCCESS)
      {
      res=hackrf_board_partid_serialno_read(devices[no_of_hackrf], &id);
      if(res == HACKRF_SUCCESS)
        {
        if( hackrf.sernum1 == id.serial_no[2] &&
            hackrf.sernum2 == id.serial_no[3])
          {
          goto open_ok;
          }
        }
      }  
    else
      {
      no_of_hackrf++;
      }
    }
  sprintf(s,"Waiting %.2f", current_time()-dt1);
  lir_text(3,5,s);
  lir_refresh_screen();
  if(kill_all_flag)goto hackrf_init_error_exit2;
  lir_sleep(300000);
  for(i=0; i< no_of_hackrf; i++)
    {
    /*!!!if(devices[no_of_hackrf] != NULL)hackrf_close(devices[i]);!!!airspy.c bug?*/
    if(devices[i] != NULL)hackrf_close(devices[i]);
    }
  no_of_hackrf=0;
  }
open_ok:;
if(no_of_hackrf > 0)
  {
  for(i=0; i< no_of_hackrf; i++)
    {
    if(devices[i] != NULL)hackrf_close(devices[i]);
    }
  }
dev_hackrf=devices[no_of_hackrf];

fprintf(dmp,"\nRate: %f\n", t1);xz("");
res=hackrf_set_samplerate(dev_hackrf, t1);
if(res != HACKRF_SUCCESS)
  {
  errcod=1404;
  goto hackrf_error_exit1;
  }
res=hackrf_set_antenna_enable(dev_hackrf, hackrf.bias_t);
if(res != HACKRF_SUCCESS)
  {
  errcod=1405;
  goto hackrf_error_exit1;
  }
res=hackrf_set_baseband_filter_bandwidth(dev_hackrf,
  (uint32_t)hackrf_bw[hackrf.bandwidth_no]*1000000);
if(res != HACKRF_SUCCESS)
  {
  errcod=2311405;
  goto hackrf_error_exit;
  }
set_hardware_rx_gain();
set_hackrf_att();
set_hackrf_frequency();
fft1_block_timing();
if(thread_command_flag[THREAD_SCREEN]!=THRFLAG_NOT_ACTIVE)
  {
  while(thread_status_flag[THREAD_SCREEN]!=THRFLAG_ACTIVE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_IDLE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_SEM_WAIT)
    {
    if(thread_command_flag[THREAD_HACKRF_INPUT] ==
                                 THRFLAG_KILL)goto hackrf_error_exit1;
    lir_sleep(10000);
    }
  }
#include "timing_setup.c"
linrad_thread_create(THREAD_HACKRF_STARTER);
thread_status_flag[THREAD_HACKRF_INPUT]=THRFLAG_ACTIVE;
lir_status=LIR_OK;
errcod=0;
while(!kill_all_flag && 
            thread_command_flag[THREAD_HACKRF_INPUT] == THRFLAG_ACTIVE)
  {
#if RUSAGE_OLD == TRUE
  if(local_workload_counter != workload_counter)
    {
    local_workload_counter=workload_counter;
    make_thread_times(THREAD_HACKRF_INPUT);
    }
#endif
  if(local_att_counter != sdr_att_counter)
    {
    local_att_counter=sdr_att_counter;
    set_hackrf_att();
    }
  if(local_nco_counter != sdr_nco_counter)
    {
    local_nco_counter=sdr_nco_counter;
    set_hackrf_frequency();
    }
  lir_await_event(EVENT_HWARE1_RXREADY);
  if(kill_all_flag)goto hackrf_error_exit;
  while (!kill_all_flag && ((timf1p_pa-timf1p_sdr+timf1_bytes)&
                                   timf1_bytemask)>snd[RXAD].block_bytes)
    {
#include "input_speed.c"
    finish_rx_read();
    if(kill_all_flag)goto hackrf_error_exit;
    }
  }
hackrf_error_exit:;
cancel_flag=TRUE;
lir_join(THREAD_HACKRF_STARTER);
hackrf_error_exit1:;
hackrf_close(dev_hackrf);
hackrf_init_error_exit2:;
hackrf_exit();/*!!!*/
unload_hackrf_library();
hackrf_init_error_exit1:;
unload_usb1_library();
hackrf_init_error_exit:;
sdr=-1;
if(errcod != 0)lirerr(errcod);
thread_status_flag[THREAD_HACKRF_INPUT]=THRFLAG_RETURNED;
while(thread_command_flag[THREAD_HACKRF_INPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
lir_close_event(EVENT_HWARE1_RXREADY);
}


int display_hackrf_parm_info(int *line)
{
char s[80];
char* sp;
int errcod;
float t1;
char *offon[2]={"OFF","ON"};
errcod=read_sdrpar(hackrf_parfil_name, MAX_HACKRF_PARM, 
                               hackrf_parm_text, (int*)((void*)&hackrf));
if(errcod == 0)
  {
  sprintf(s,"Serial no          = 0x%08X%08X",
                                         hackrf.sernum1, hackrf.sernum2);
  lir_text(24,line[0],s);
  SNDLOG"\n%s",s);
  line[0]++;
  t1 = hackrf.ratenum/1000000;
  sprintf(s,"Sampling rate      = %.2f Mhz, Xtal adjust = %.0f ppb",t1,
                                                   10.*hackrf.freq_adjust);
  lir_text(24,line[0],s);
  SNDLOG"\n%s",s);
  line[0]++;
  if(hackrf.gain_mode == 0)
    {
    sp="Linearity";
    }
  else
    {
    sp="Sensitivity";
    }
  sprintf(s,"Gain mode          = %s, Bandwidth = %.1f MHz",sp,
                                           hackrf_bw[hackrf.bandwidth_no]);
  lir_text(24,line[0],s);
  SNDLOG"\n%s",s);
  line[0]++;
  sprintf(s,"Bias T (5V)        = %s",offon[hackrf.bias_t]);
  lir_text(24,line[0],s);
  SNDLOG"\n%s",s);
  }
return errcod;
}

void init_hackrf(void)
{
FILE *hackrf_file;
const char cind[4]={'-','/','|','\\'};
struct hackrf_device* devices[MAX_HACKRF_DEVICES];
hackrf_read_partid_serialno_t id[MAX_HACKRF_DEVICES];
const char *name[MAX_HACKRF_DEVICES];
char ver[MAX_HACKRF_DEVICES][25];
uint8_t nid;
int no_of_hackrf, devno, devnum[MAX_HACKRF_DEVICES];
hackrf_error res;
int i, k, m, line;
int *sdr_pi;
double t1;
char s[80];

clear_screen();
load_usb1_library(TRUE);
if(libusb1_library_flag == FALSE)return;
load_hackrf_library();
if(hackrf_library_flag == FALSE){unload_usb1_library();return;}

no_of_hackrf=0;
line=1;
k=0;
hackrf_init();/*!!!*/
res=HACKRF_ERROR_NOT_FOUND;
while( (res != HACKRF_SUCCESS && no_of_hackrf == 0) ||
       (res == HACKRF_SUCCESS && no_of_hackrf < MAX_HACKRF_DEVICES) )
  {
  res=hackrf_open(&devices[no_of_hackrf]);
  if(res == HACKRF_SUCCESS)
    {
    no_of_hackrf++;
    }
  else
    {
    if(no_of_hackrf == 0)
      {
      sprintf(s,"%s  %c", hackrf_error_name(res), cind[k]);
      lir_text(5,line,s);
      lir_refresh_screen();
      lir_sleep(100000);
      test_keyboard();
      if(kill_all_flag || lir_inkey != 0)
        {
        goto unload_lib;
        }
      k++;
      k&=3;
      }
    }
  }
clear_lines(line,line);
for(k=0; k<no_of_hackrf; k++)
  {
  if ( hackrf_board_id_read(devices[k], &nid) != HACKRF_SUCCESS
  ||   hackrf_board_partid_serialno_read(devices[k], &id[k]) != HACKRF_SUCCESS
  ||   hackrf_version_string_read(devices[k], ver[k], sizeof(ver[0]))
         != HACKRF_SUCCESS )
    {
    hackrf_close(devices[k]);
    devices[k]=NULL;
    }
  name[k]=hackrf_board_id_name(nid);
  }
m=0;
// Windows XP will open the same device multiple times so
// a second device will not be opened. Make sure we just close it once.
if( no_of_hackrf > 1 && 
    id[0].serial_no[2] == id[1].serial_no[2] && 
    id[0].serial_no[3] == id[1].serial_no[3])
  {
  no_of_hackrf=1;
  }
for(k=0; k<no_of_hackrf; k++)
  {
  if(devices[k])
    {
    snprintf(s, sizeof(s), "%2d: serial no = 0x%08X%08X  %s  %s",
             m, id[k].serial_no[2],id[k].serial_no[3],name[k],ver[k]);
    lir_text(1,line,s);
    line++;
    devnum[m]=k;
    m++;
    }
  }
line++;
if(m == 0)
  {
  lir_text(5,line,"No usable HackRf device found");
errexit:;
  await_processed_keyboard();
  hackrf_exit();/*!!!*/
  unload_hackrf_library();
  unload_usb1_library();
  return;
  }
else
  {
  if(m == 1)
    {
    lir_text(5,line,"HackRf device autoselected");
    devno=devnum[0];
    }
  else
    {
    lir_text(5,line,"Select device by line number.");
    devno=devnum[lir_get_integer(34,line,2,0,m-1)];
    if(kill_all_flag)goto unload_lib;
    }
  }
hackrf.sernum1=id[devno].serial_no[2];
hackrf.sernum2=id[devno].serial_no[3];
dev_hackrf=devices[devno];

line+=3;
get_bias_t:;
lir_text(2,line,"Enable bias T (Y/N)?");
await_processed_keyboard();
if(kill_all_flag)goto unload_lib;
if(lir_inkey == 'Y')
  {
  hackrf.bias_t=1;
  lir_text(0,line+2,
      "WARNING!! You have selected to send 5V DC to the input connector");
  lir_text(0,line+3,"The current can be 0.5 A.");
  lir_text(0,line+4,"Do not connect signal generators etc.");
  lir_text(0,line+6,"Do you really want this? (Y/N)");
confirm:;
  await_processed_keyboard();
  if(kill_all_flag)goto unload_lib;
  if(lir_inkey == 'Y')
    {
    clear_lines(line,line+6);
    lir_text(5,line,"Enable bias T");
    }
  else
    {
    if(lir_inkey != 'N')goto confirm;
    clear_lines(line,line+6);
    goto get_bias_t;
    }
  }
else
  {
  if(lir_inkey != 'N' && lir_inkey != '0')goto get_bias_t;
  clear_lines(line,line);
  line-=3;
  hackrf.bias_t=0;
  }
res = hackrf_set_antenna_enable(dev_hackrf, hackrf.bias_t);
if(res != HACKRF_SUCCESS)
  {
  sprintf(s,"Failed to set bias T: %s", hackrf_error_name(res));
  line+=2;
  lir_text(2,line,s);
  goto errexit_close;
  }
#define FIRSTSRKEY '0'
m = line+=3;
for(i=0; i<MAX_HACKRF_SR; i++)
  {
  sprintf(s,"%c => %.1f MHz",FIRSTSRKEY+(char)i,TOMD(hackrf_sr[i]));
  lir_text(5,line,s);
  line++;
  }
line++;
lir_text(2,line,"Select samplerate from table.");
getsr:;
await_processed_keyboard();
if(kill_all_flag)goto unload_lib;
i=lir_inkey-FIRSTSRKEY;
if(i < 0)goto getsr;
if(i >= MAX_HACKRF_SR)goto getsr;
hackrf.ratenum=hackrf_sr[i];
clear_lines(m,line);
line = m;
sprintf(s,"Selected samplerate is %.1f MHz",TOMD(hackrf_sr[i]));
lir_text(5,line,s);
res=hackrf_set_samplerate(dev_hackrf, hackrf.ratenum);
if(res != HACKRF_SUCCESS)
  {
  sprintf(s,"Failed to set sample rate: %s", hackrf_error_name(res));
  lir_text(2,line,s);
errexit_close:;
  for(k=0; k<no_of_hackrf; k++)
    {
    if(devices[k])
      {
      hackrf_close(devices[k]);
      }
    }
  goto errexit;
  }
m = line+= 3;
#define FIRSTBWKEY '0'
for(i=0; i<MAX_HACKRF_BW; i++)
  {
  sprintf(s,"%c => %.1f MHz",FIRSTBWKEY+(char)i,hackrf_bw[i]);
  lir_text(5,line,s);
  line++;
  }
line++;
lir_text(2,line,"Select bandwidth from table.");
getbw:;
await_processed_keyboard();
if(kill_all_flag)goto unload_lib;
i=lir_inkey-FIRSTBWKEY;
if(i < 0)goto getbw;
if(i >= MAX_HACKRF_BW)goto getbw;
hackrf.bandwidth_no=(unsigned int)i;
clear_lines(m,line);
line = m;
sprintf(s,"Selected bandwidth  is %.1f MHz",hackrf_bw[i]);
lir_text(5,line,s);
line +=3;
lir_text(3,line,"Select gain mode.");
line++;
lir_text(5,line,"0 => Linearity.");
line++;
lir_text(5,line,"1 => Sensitivity.");
line+=2;
lir_text(3,line,"=>");
hackrf.gain_mode=lir_get_integer(6,line,1,0,1);
line+=3;
t1 = hackrf.ratenum;
hackrf.freq_adjust=0;
#if ADJ_HACKRF_XTAL
lir_text(3,line,"Enter xtal error in ppb =>");
hackrf.freq_adjust=0.1*lir_get_float(32,line,9,-300000.,300000.);
if(kill_all_flag)goto unload_lib;
t1=((rint)((t1*(100000000L+(double)hackrf.freq_adjust))/100000000L)); /*it is rint()*/
#endif
/*ui.rx_ad_speed=hackrf.ratenum=(unsigned int)t1;*/
ui.rx_ad_speed=(unsigned int)t1;

hackrf_file=fopen(hackrf_parfil_name,"w");
if(hackrf_file == NULL)
  {
  lirerr(382368);
  goto unload_lib;
  }
hackrf.check=HACKRF_PAR_VERNR;
sdr_pi=(int*)(&hackrf);
for(i=0; i<MAX_HACKRF_PARM; i++)
  {
  fprintf(hackrf_file,"%s [%d]\n",hackrf_parm_text[i],sdr_pi[i]);
  }
parfile_end(hackrf_file);
ui.rx_addev_no=HACKRF_DEVICE_CODE;
ui.rx_rf_channels=1;
ui.rx_input_mode=IQ_DATA;
ui.rx_ad_channels=2;
ui.rx_admode=0;
unload_lib:;
m=0;
for(k=0; k<no_of_hackrf; k++)
  {
  if(devices[k])
    {
    res=hackrf_close(devices[k]);
    if(res != HACKRF_SUCCESS)m++;
    }
  }
hackrf_exit();
unload_hackrf_library();
unload_usb1_library();
if(m != 0)
  {
  lirerr(1315);
  return;
  }
}
